@@ -8,6 +8,8 @@
  * memorandum "A Kernel Model for Precision Timekeeping" by Dave Mills,
  * revised 31 January 1996. That document updates RFC1589.
  *
+ * This program has been updated to use the new PPS API
+ *
  * Copyright (c) 1996 - 1999 by Ulrich Windl
  *
  * This program is free software; you can redistribute it and/or
@@ -24,7 +26,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * $Id: enable_pps.c,v 1.20 1999/04/24 19:10:59 windl Exp $
+ * $Id: enable_pps.c,v 1.25 1999/05/23 19:22:04 windl Exp $
  */
 #include	<unistd.h>
 #include	<time.h>
@@ -39,9 +41,10 @@
 #include	<sys/timex.h>
 
 #include	<sys/ppsclock.h>
-#include	<linux/serial.h>
+
+#include	<sys/timepps.h>
 #if defined(__GNU_LIBRARY__) && __GNU_LIBRARY__ >= 6	/* libc6 or glibc2 */
-# warning	"Compatibility with this library has not been tested a lot!"
+/*# warning	"Compatibility with this library has not been tested a lot!" */
 # include	<ioctl-types.h>
 #else
 # include	<sys/ioctl.h>
@@ -71,7 +74,7 @@
 #define ADJ_ADJTIME	ADJ_OFFSET_SINGLESHOT	/* compatibility define */
 #endif
 
-#define OFFSET_TOLERANCE	333333		/* us */
+#define OFFSET_TOLERANCE	333333333	/* ns */
 #define STABIL_TOLERANCE	(10 << SHIFT_USEC)
 #define	PPS_INSANE		(STA_PPSERROR|STA_PPSWANDER|STA_PPSJITTER)
 #define ALL_ERRORS		(PPS_INSANE|STA_CLOCKERR)
@@ -81,7 +84,7 @@
 #define	BITS_OFF(v, b)			((v) & ~(b))
 #define	BITS_ON(v, b)			((v) | (b))
 
-static	const char	id[] = "$Id: enable_pps.c,v 1.20 1999/04/24 19:10:59 windl Exp $";
+static	const char	id[] = "$Id: enable_pps.c,v 1.25 1999/05/23 19:22:04 windl Exp $";
 
 #define	BIT(n)	(1UL << (n))
 #define OPT_MONITOR		BIT(0)	/* perform monitor loop */
@@ -96,16 +99,23 @@ static	const char	id[] = "$Id: enable_pps.c,v 1.20 1999/04/24 19:10:59 windl Exp
 
 static	unsigned long	options = OPT_MONITOR | OPT_RESET;
 
-static	char	polarity	= 'p';	/* p[ositive] | n[egative] | 0) */
+static	char	trigger	= 'A';		/* off | assert | clear | both */
 
 static	const	int	fd	= 0;	/* file descriptor of port (stdin) */
+static	pps_handle_t	ppshandle;	/* PPS handle (stdin) */
 
-struct clock_info {		/* information about clock and conditions */
-	struct timex		tx;
-	struct ppsclockev	ev;
-	struct timeval		delta;	/* time difference sincle pulses */
-	long			jitter;	/* jitter between pulses (us) */
-	int			state;	/* time state */
+struct clock_info {			/* information about clock and conditions */
+	struct timex	tx;		/* kernel status information */
+	pps_info_t	info;		/* pulse information */
+	struct timespec	delta;		/* time difference between pulses */
+	int		have_assert;	/* assert event detected */
+	int		have_clear;	/* clear event detected */
+	long		sec, nsec;	/* last time stamp */
+	long		dsec, dnsec;	/* difference since last time stamp */
+	long		seq, dseq;	/* sequence number and difference */
+	long		nano_offset;	/* offset around zero (ns) */
+	long		jitter;		/* jitter between pulses (ns) */
+	int		state;		/* time state */
 };
 
 typedef struct {		/* general monitor */
@@ -194,21 +204,20 @@ static	int	pps_status(const char *tag, const int show_event,
 	char const	*state_str = NULL;
 	char const	*sep	= tag;
 
-	if ( show_event && cip->ev.serial != last.ev.serial )
+	if ( show_event && (cip->have_assert || cip->have_clear) )
 	{
-		long	secs	= cip->ev.tv.tv_sec - last.ev.tv.tv_sec;
-		long	evs	= cip->ev.serial - last.ev.serial;
+		long	sec = cip->sec;
 
-		if ( cip->ev.tv.tv_usec - last.ev.tv.tv_usec < -500000 )
-			--secs;
-		else if ( cip->ev.tv.tv_usec - last.ev.tv.tv_usec > 500000 )
-			++secs;
-		fprintf(stderr, "%spulse %u at 0.%06d",
-			sep, cip->ev.serial, cip->ev.tv.tv_usec), sep = ", ";
-		if ( secs != evs )
+		if ( cip->nsec < -500000000 )
+			--sec;
+		else if ( cip->nsec > 500000000 )
+			++sec;
+		fprintf(stderr, "%spulse(%c:%c) %lu at 0.%09ld",
+			sep, cip->have_assert + '0', cip->have_clear + '0',
+			cip->seq, cip->nsec), sep = ", ";
+		if ( sec != cip->dseq )
 			fprintf(stderr, "%s%ld events in %ld seconds",
-				sep, evs, secs), sep = ", ";
-		last.ev = cip->ev;
+				sep, cip->dseq, sec), sep = ", ";
 		result = 0;
 	}
 	switch ( cip->state )
@@ -221,9 +230,6 @@ static	int	pps_status(const char *tag, const int show_event,
 	}
 	if ( state_str != NULL )
 		fprintf(stderr, "%sstate=%s", sep, state_str), sep = ", ";
-
-	if ( cip->ev.serial == 0 || !(cip->tx.status & STA_PPSSIGNAL) )
-		fprintf(stderr, "%sno pulse", sep), sep = ", ";
 	if ( ((cip->tx.status ^ last.tx.status) & ~ALL_ERRORS) != 0 )
 	{
 		fprintf(stderr, "%snew status=", sep), sep = ", ";
@@ -243,37 +249,64 @@ static	int	pps_status(const char *tag, const int show_event,
 /* update *cip */
 static	int	update_clock_info(struct clock_info *cip)
 {
-	struct clock_info	last = *cip;
+	static struct clock_info	last;
 	int	result = -1;
 
 	cip->tx.modes = 0;
 	cip->state = adjtimex(&cip->tx);
-	if ( ioctl(fd, CIOGETEV, &cip->ev) == -1 )
+	if ( time_pps_fetch(ppshandle, &cip->info) == -1 )
 	{
-		perror("ioctl(CIOGETEV)");
-		memset(&cip->ev, 0, sizeof(cip->ev));
+		perror("time_pps_fetch()");
+		memset(&cip->info, 0, sizeof(cip->info));
 	}
 	
-	if ( cip->ev.serial != last.ev.serial )
+	cip->have_assert = cip->info.assert_sequence != last.info.assert_sequence;
+	cip->have_clear = cip->info.clear_sequence != last.info.clear_sequence;
+	if ( cip->have_assert )
 	{
-		long	dsecs	= cip->ev.tv.tv_sec - last.ev.tv.tv_sec;
-		long	dusec	= cip->ev.tv.tv_usec - last.ev.tv.tv_usec;
-
-		cip->delta.tv_sec = dsecs;
-		cip->delta.tv_usec = dusec;
-		if ( dusec < -500000 )
-			--dsecs;
-		else if ( dusec > 500000 )
-			++dsecs;
-		cip->jitter = dsecs * 1000000 + dusec;
-		cip->jitter -= last.delta.tv_sec * 1000000;
-		cip->jitter -= last.delta.tv_usec;
+		cip->sec = cip->info.assert_timestamp.tv_sec;
+		cip->dsec = cip->sec - last.sec;
+		cip->nsec = cip->info.assert_timestamp.tv_nsec;
+		cip->dnsec = cip->nsec - last.nsec;
+		cip->seq = cip->info.assert_sequence;
+		cip->dseq = cip->seq - last.seq;
+	}
+	if ( cip->have_clear )
+	{
+		cip->sec = cip->info.clear_timestamp.tv_sec;
+		cip->dsec = cip->sec - last.sec;
+		cip->nsec = cip->info.clear_timestamp.tv_nsec;
+		cip->dnsec = cip->nsec - last.nsec;
+		cip->seq = cip->info.clear_sequence;
+		cip->dseq = cip->seq - last.seq;
+	}
+	if ( cip->dseq )
+	{
+		cip->nano_offset = cip->nsec;
+		if ( cip->nano_offset > 500000000 )
+			cip->nano_offset -= 1000000000;
+		cip->delta.tv_sec = cip->dsec;
+		cip->delta.tv_nsec = cip->dnsec;
+		if ( cip->dnsec < -500000000 )
+		{
+			--cip->delta.tv_sec;
+			cip->delta.tv_nsec += 1000000000;
+		}
+		else if ( cip->dnsec > 500000000 )
+		{
+			++cip->delta.tv_sec;
+			cip->delta.tv_nsec -= 1000000000;
+		}
+		cip->jitter = cip->delta.tv_sec - last.delta.tv_sec;
+		cip->jitter *= 1000000000;
+		cip->jitter += cip->delta.tv_nsec - last.delta.tv_nsec;
+		last = *cip;
 		result = 0;
 	}
 	return(result);
 }
 
-/* do adjtimex and check for errors */
+/* do adjtimex(), update clock_info, and check for errors */
 static	int	verbose_adjtimex(const char *tag, const int show_event,
 				 struct clock_info *cip)
 {
@@ -364,28 +397,34 @@ static	void	update_constant(struct clock_info *cip)
 static	void	pll_adj_offset(struct clock_info *cip)
 {
 	static	time_t last_adj_time;
-	static	u_int  last_serial;
 	time_t	t = time(NULL);
-	long	offset = cip->ev.tv.tv_usec;
+	long	offset = 0;
 	long	tdelta, tconst;
 
-	if ( last_serial == cip->ev.serial )
+	if ( cip->dseq == 0 )
 		return;
-	last_serial = cip->ev.serial;
-	if ( offset > 500000 )
-		offset -= 1000000;	/* change direction */
+	if ( cip->have_assert )
+		offset = cip->info.assert_timestamp.tv_nsec;
+	if ( cip->have_clear )
+		offset = cip->info.clear_timestamp.tv_nsec;
+	if ( offset > 500000000 )
+		offset -= 1000000000;	/* change direction */
 	offset = -offset;		/* correct direction */
 	cip->tx.modes = MOD_MAXERROR | MOD_ESTERROR;
-	cip->tx.esterror = cip->tx.maxerror = offset >= 0 ? : -offset;
 	cip->tx.offset = offset;
+	offset /= 1000;			/* convert ns to us */
+	cip->tx.esterror = cip->tx.maxerror = offset >= 0 ? : -offset;
 	if ( cip->tx.offset > MAXPHASE / 2 || cip->tx.offset < -MAXPHASE / 2 )
+	{
 		cip->tx.modes = ADJ_ADJTIME;
+		cip->tx.offset = offset;	/* need us here */
+	}
 	else
 	{
 		cip->tx.modes |= MOD_OFFSET;
 #ifdef STA_NANO
-		if ( (cip->tx.status & STA_NANO) != 0 )
-			cip->tx.offset *= 1000;
+		if ( (cip->tx.status & STA_NANO) == 0 )
+			cip->tx.offset /= 1000;
 #endif
 	}
 	tdelta = t - last_adj_time;
@@ -419,21 +458,16 @@ static	int	trg_offset(const struct clock_info *cip)
 
 static	void	act_offset(struct clock_info *cip)
 {
-	static	struct clock_info	last;
-	long offset = cip->ev.tv.tv_usec;
+	long offset = cip->nano_offset;
 
-	if ( offset > 500000 )
-		offset -= 1000000;	/* round to nearest second */
 	if ( offset < -OFFSET_TOLERANCE || offset > OFFSET_TOLERANCE )
 		update_status(cip, 0, STA_FREQHOLD | STA_PLL);
 	else if ( offset > -OFFSET_TOLERANCE / 2 &&
 		  offset < OFFSET_TOLERANCE / 2 )
 		update_status(cip, STA_FREQHOLD, STA_PLL);
-	if ( (cip->tx.status & STA_PPSSIGNAL) != 0 &&
-	     cip->ev.serial != last.ev.serial )
+	if ( (cip->tx.status & STA_PPSSIGNAL) != 0 && cip->dseq != 0 )
 	{
 		pll_adj_offset(cip);
-		last = *cip;
 	}
 	update_constant(cip);
 }
@@ -448,10 +482,10 @@ static	int	trg_freq(const struct clock_info *cip)
 
 static	void	act_freq(struct clock_info *cip)
 {
-	long offset = cip->ev.tv.tv_usec;
+	long offset = cip->dnsec;
 
-	if ( offset > 500000 )
-		offset -= 1000000;	/* round to nearest second */
+	if ( offset > 500000000 )
+		offset -= 1000000000;	/* round to nearest second */
 	if ( offset < -OFFSET_TOLERANCE || offset > OFFSET_TOLERANCE )
 		update_status(cip, STA_FLL, STA_FREQHOLD | STA_PLL);
 	else if ( offset > -OFFSET_TOLERANCE / 2 &&
@@ -469,10 +503,8 @@ static	int	trg_ppstime(const struct clock_info *cip)
 
 static	void	act_ppstime(struct clock_info *cip)
 {
-	long offset = cip->ev.tv.tv_usec;
+	long offset = cip->nano_offset;
 
-	if ( offset > 500000 )
-		offset -= 1000000;	/* round to nearest second */
 	if ( (cip->tx.status & STA_PPSSIGNAL) != 0
 	     && (cip->tx.status & PPS_INSANE) == 0
 	     && (offset > -OFFSET_TOLERANCE && offset < OFFSET_TOLERANCE) )
@@ -532,9 +564,10 @@ static	int	trg_wait(const struct clock_info *cip)
 
 static	void	act_wait(struct clock_info *cip)
 {
-	static struct timespec	sleeper = {0, 500000000};
+	static struct timespec	sleeper = {5, 0};
+	pps_info_t		buf;
 
-	nanosleep(&sleeper, &sleeper);
+	time_pps_wait(ppshandle, &sleeper, &buf);
 }
 
 static monitor_t	monitors[] = {
@@ -566,26 +599,24 @@ static	void	calibrate(struct clock_info *cip)
 static	void	monitor_jitter(struct clock_info *cip)
 {
 	struct clock_info	last;
-	struct timespec	sleeper;
+	pps_info_t		buf;
 
 	cip->tx.modes = 0;
-	sleeper.tv_sec = 0;
-	sleeper.tv_nsec = 500000000;
 	while ( 1 )
 	{
+		time_pps_wait(ppshandle, NULL, &buf);
 		update_clock_info(cip);
-		if ( cip->ev.serial == last.ev.serial )
-		{
-			nanosleep(&sleeper, &sleeper);
+		if ( cip->seq == last.seq )
 			continue;
-		}
 		pps_status("monitor_jitter(): ", 0, cip);
-		printf("%6d %8ld", cip->ev.serial, cip->jitter);
-		if ( cip->jitter > 1000 || cip->jitter < -1000 )
+		printf("event(%4lu:%4lu) jitter %9ld ns, offset %9ld ns",
+		       cip->info.assert_sequence, cip->info.clear_sequence,
+		       cip->jitter, cip->nano_offset);
+		if ( cip->jitter > 1000000 || cip->jitter < -1000000 )
 		{
-			printf("\t%d:%ld %d:%ld",
-			       last.ev.tv.tv_sec, last.ev.tv.tv_usec,
-			       cip->ev.tv.tv_sec, cip->ev.tv.tv_usec);
+			printf("\t%ld:%ld %ld:%ld %ld:%ld",
+			       last.sec, last.nsec, cip->sec, cip->nsec,
+			       cip->delta.tv_sec, cip->delta.tv_nsec);
 		}
 		last = *cip;
 		printf("\n");
@@ -608,8 +639,8 @@ static	void	usage(void)
 		"\t-N\tenforce MOD_NANO\n"
 #endif
 		"\t-O\tdon't adjust offset in monitoring loop\n"
-		"\t-p{p|0|n}\tpolarity (of pulse) positive|none|negative\n"
 		"\t-S\tdon't adjust status in monitoring loop\n"
+		"\t-t{0|a|c|b}[h]\ttrigger for event: off|assert|clear|both\n"
 		);
 }
 
@@ -618,8 +649,8 @@ int	main(int argc, char *argv[])
 {
 	struct utsname	un;
 	int		stat = TIOCM_RTS;	/* RTS MODEM control line */
-	struct serial_struct ss;
 	struct clock_info ci;
+	pps_params_t	parm;
 	int		ch;
 
 	fprintf(stderr,
@@ -631,7 +662,7 @@ int	main(int argc, char *argv[])
 	       " nanokernel support.\n", UTS_RELEASE);
 	uname(&un);
 	printf(" Now running %s %s)\n", un.sysname, un.release);
-	while ( (ch = getopt(argc, argv, "CeFjkMNOp:S")) != -1 )
+	while ( (ch = getopt(argc, argv, "CeFjkMNOSt:")) != -1 )
 	{
 		switch ( ch )
 		{
@@ -660,24 +691,26 @@ int	main(int argc, char *argv[])
 		case 'O':	/* don't adjust offset */
 			options = BITS_ON(options, OPT_NO_OFFSET);
 			break;
-		case 'p':	/* select desired edge of the pulse */
-			polarity = optarg[0];
-			switch ( polarity )
-			{
-			case 'p':
-			case 'n':
-			case 'm':
-			case '0':
-				break;
-			default:
-				fprintf(stderr, "polarity '%c' is illegal\n",
-					polarity);
-				polarity = '0';
-			}
-			break;
 		case 'S':	/* don't adjust time status */
 			options = BITS_ON(options, OPT_NO_STATUS);
 			break;
+		case 't':	/* select desired edge of the pulse */
+			trigger = optarg[0];
+			switch ( trigger )
+			{
+			case '0':
+			case 'a':
+			case 'c':
+			case 'b':
+				break;
+			default:
+				fprintf(stderr, "trigger '%s' is illegal\n",
+					optarg);
+				trigger = '0';
+			}
+			if ( optarg[1] == 'h' )
+				trigger = toupper(trigger);
+			break;
 		case '?':
 			fprintf (stderr, "Unknown option `-%c'.\n", optopt);
 			usage();
@@ -692,25 +725,33 @@ int	main(int argc, char *argv[])
 	}
 
 	ioctl(fd, TIOCMBIC, &stat);	/* power-on DCF77 receiver */
-	ioctl(fd, TIOCGSERIAL, &ss);	/* enable pulse detection on DCD */
-	switch ( polarity )
+	if ( time_pps_create(fd, &ppshandle) )
+		perror("time_pps_create()");
+	parm.mode = 0;			/* clear all flags */
+	switch ( trigger )
 	{
-	case '0':
-		ss.flags &= ~(ASYNC_PPS_CD_POS | ASYNC_PPS_CD_NEG); break;
-	case 'p':
-		ss.flags &= ~ASYNC_PPS_CD_NEG;
-		ss.flags |= ASYNC_PPS_CD_POS;
+	case 'A':
+		parm.mode |= PPS_HARDPPSONASSERT;	/* fall through */
+	case 'a':
+		parm.mode |= PPS_CAPTUREASSERT;
+		break;
+	case 'C':
+		parm.mode |= PPS_HARDPPSONCLEAR;	/* fall through */
+	case 'c':
+		parm.mode |= PPS_CAPTURECLEAR;
+		break;
+	case 'B':
+		/* parm.mode |= PPS_HARDPPSONBOTH; */	/* fall through */
+	case 'b':
+		parm.mode |= PPS_CAPTUREBOTH;
 		break;
-	default:
-		ss.flags &= ~ASYNC_PPS_CD_POS;
-		ss.flags |= ASYNC_PPS_CD_NEG;
 	}
-	if ( ioctl(fd, TIOCSSERIAL, &ss) != 0 )
-		perror("ioctl(TIOCSSERIAL) to enable PPS detection");
+	if ( time_pps_setparams(ppshandle, &parm) != 0 )
+		perror("time_pps_setparams()");
 
 	fprintf(stderr,
-		"PPS detection enabled on CD-pin for polarity '%c'\n",
-		polarity);
+		"PPS API set up to detect '%c' events\n",
+		trigger);
 	memset(&ci, 0, sizeof(ci));
 	update_clock_info(&ci);
 	print_header();
